package org.lym.pom.repository.mongo;

import lombok.AllArgsConstructor;
import lombok.Data;
import lombok.NoArgsConstructor;
import org.springframework.data.annotation.Id;
import org.springframework.data.mongodb.core.index.Indexed;
import org.springframework.data.mongodb.core.mapping.Document;
import org.springframework.data.mongodb.core.mapping.Field;

import java.io.Serializable;

/**
 * 一个用户对另一个用户的状态/评论进行互动的载体，如赞、喜欢、反对、评论/回复
 */
@Data
@AllArgsConstructor
@NoArgsConstructor
@Document(collection = "comment")
public class Comment implements Serializable {

    /**
     * 评论 id
     */
    @Id
    private String id;

    /**
     * 所属对象 id，可以是文章的
     * 评论的评论，该值仍然为被评论对象id
     */
    @Indexed
    private String relatedId;
    /**
     * 所属对象类型，如文章、动态、视频等，冗余信息用于扩展与提高读性能
     */
    @Field
    private String relatedType;
    /**
     * 所属对象的创建人id，冗余信息用于提高读性能
     * 评论通知触发后，给对应人发被评论消息通知使用，减少查库
     */
    @Field
    private String relatedUserId;
    // ------------------

    /**
     * 父节点id，默认0
     */
    @Indexed
    private String parentId;
    /**
     * 父节点的创建人id，冗余信息用于提高读性能
     * 评论通知触发后，给对应人发被评论消息通知使用，减少查库
     */
    @Field
    private String parentUserId;

    /**
     * 类型：评论、赞、喜欢
     */
    @Field
    private String type;

    /**
     * 内容
     */
    @Field
    private String content;
    /**
     * 状态
     * 待审核，不可见
     * 正常可见
     * 管理移除，不可见
     * 用户删除，不可见
     */
    @Field
    private String status;

    /**
     * 推荐评分（推荐系统打分，便于推荐优秀的评论）
     */
    @Indexed
    private Double score;

    /**
     * 评论人
     */
    @Field
    private String creator;

    /**
     * 创建时间
     */
    @Field
    private Long createTime;

    /**
     * 创建时间
     */
    @Field
    private Long updateTime;

    /**
     * 互动对象类型
     * 用户状态/分享
     * 视频
     * 照片
     */
    public interface RelatedType {
        String MOMENT = "M";
        String VIDEO = "V";
        String PHOTO = "P";
    }

    /**
     * 互动类型：评论、赞、喜欢、反对、举报
     */
    public interface Type {
        String LIKE = "L";
        String LOVE = "LO";
        String DISLIKE = "D";
        String COMMENT = "C";
//        String REPORT = "R";
    }

    /**
     * 状态
     * 正常可见
     * 待作者精选，仅作者和发表人可见
     * 待审核，不可见
     * 管理移除，不可见
     * 用户删除，不可见
     */
    public interface Status {
        String NORMAL = "N";
        String REVIEW = "R";
        String AUDIT = "A";
        String DELETED = "D";
        String FORCE_REMOVE = "FR";
    }
}
