package org.lym.pom.service.impl;

import org.lym.pom.notify.publisher.EventPublisher;
import org.lym.pom.repository.mongo.Comment;
import org.lym.pom.service.CommentService;
import org.shoulder.core.exception.CommonErrorCodeEnum;
import org.shoulder.core.util.AssertUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.stereotype.Service;

/**
 * 订阅推送服务
 *
 * @author lym
 */
@Service
public class CommentServiceImpl implements CommentService {


    @Autowired
    private EventPublisher eventPublisher;


    @Autowired
    private MongoTemplate mongoTemplate;

    @Override
    public boolean addComment(Comment comment) {
        // todo 进阶功能 检查被评论的是否开启禁止评论等

//        comment.setId("todo");
        comment.setType(Comment.Type.COMMENT);
        comment.setStatus(Comment.Status.NORMAL);
        comment.setScore(10.0);
        comment.setCreateTime(System.currentTimeMillis());
        comment.setUpdateTime(System.currentTimeMillis());
        mongoTemplate.save(comment);
        return true;
    }

    /**
     * 删除评论，必须根据 id 删除，避免多条评论都删了
     * @param comment 评论信息 id
     */
    @Override
    public void removeComment(Comment comment) {
        // 调用方校验是自己还在可删的评论
        AssertUtils.equals(Comment.Type.COMMENT, comment.getType(), CommonErrorCodeEnum.ILLEGAL_PARAM);
        AssertUtils.notEmpty(comment.getId(), CommonErrorCodeEnum.ILLEGAL_PARAM);
        mongoTemplate.remove(comment);
    }

    @Override
    public boolean addMarkedComment(Comment comment) {
        AssertUtils.notEquals(comment.getType(), Comment.Type.COMMENT, CommonErrorCodeEnum.ILLEGAL_PARAM);
        Criteria criteria = Criteria.where("type").is(comment.getType())
                .and("relatedId").is(comment.getRelatedId())
                .and("relatedType").is(comment.getRelatedType())
                .and("parentId").is(comment.getParentId())
                .and("creator").is(comment.getCreator());
        Query query = new Query(criteria);
        boolean exists = mongoTemplate.exists(query, Comment.class);
        if (exists) {
            // 幂等
            return true;
        }

        comment.setContent(null);
        comment.setStatus(Comment.Status.NORMAL);
        comment.setScore(0.0);
        comment.setCreateTime(System.currentTimeMillis());
        comment.setUpdateTime(System.currentTimeMillis());
        mongoTemplate.save(comment);
        return true;
    }

    /**
     * 移除赞，最多一个赞，根据挂载节点删，就算add时并发加多了，这里可以全删
     * @param comment 评论信息
     */
    @Override
    public void removeMarkedComment(Comment comment) {
        AssertUtils.notEquals(comment.getType(), Comment.Type.COMMENT, CommonErrorCodeEnum.ILLEGAL_PARAM);
        Query query = Query.query(Criteria.where("relatedId").is(comment.getRelatedId())
                .and("relatedType").is(comment.getType())
                .and("creator").is(comment.getCreator())
                .and("parentId").is(comment.getParentId())
                .and("type").is(comment.getType())
        );
        mongoTemplate.remove(query, Comment.class);
    }

    @Override
    public long countCommentNumber(String relatedId, String relatedType, String parentId, String type) {
        Criteria criteria = Criteria.where("relatedId").is(relatedId)
                .and("relatedType").is(relatedType)
                .and("type").is(type);
        if(parentId != null){
            criteria = criteria.and("parentId").is(parentId);
        }
        Query query = Query.query(criteria);
        return mongoTemplate.count(query, Comment.class);
    }
}
